;Spc2Midi - Opcode Disassembly
;============================================================

section .text
;==============================
; Shows a disassembled opcode line or displays a graphical path of the PC
;
;(esi=opcode address, bx=A:Y, al=flags)
ShowOpcode:
.Name:
    pusha
    mov [SpcEmu.AY],bx
    mov [SpcEmu.Pc],si
    mov [SpcEmu.Flags],al
    mov edi,MakeOperandString.Buffer
    call MakeOperandString
    mov edx,MakeOperandString.Buffer
    call WriteString
    popa
    ret

%if SpcDebugMode=2
.Path:
	;push eax
	mov ah,[.PathTable+esi]
	inc ah
    or ah,128
	;push ds
	;mov es,[AbsoluteSelector]
	;mov [es:0A0000h+esi],ah
.NoTableOverflow:
	mov [.PathTable+esi],ah
	mov [0A0000h+esi],ah
	;pop es
	;pop eax
	ret

.PathInit:
	mov edi,.PathTable
	mov ecx,65536/4
	xor eax,eax
	cld
	rep stosd

	mov eax,13h
	int 10h
	xor eax,eax
	mov dx,3c8h
	out dx,al
	inc edx
    xor ecx,ecx
.NextGrayColor:
    out dx,al
    out dx,al
    out dx,al
	add ah,64
	adc al,0
	dec cl
	jnz .NextGrayColor
	mov edi,0A0000h
	mov eax,4030201h
	mov ecx,64

.NextColorBarSection:
	stosd
	add eax,4040404h
	loop .NextColorBarSection
	ret

section .bss
alignb 4
.PathTable:     resb 65536
section .text
%endif

%if SpcDebugMode=3
.SetVideoMode:
    mov eax,13h
    int 10h
    ret
%endif

;==============================
;(edi=destination buffer)
;turns the current opcode and its 0 to 2 operands into a mnemonic string
;also shows the current PC,A,Y,X,Sp and flag values
MakeOperandString:
.MaxOpcIndent   equ 15
	cld

	mov esi,[SpcEmu.Pc]
	mov edx,[esi+SpcRam]            ;get opcode and any operands
    mov [.Opcode],edx
    movzx ebx,dl                    ;get opcode value
    mov eax,[SpcOpcInfo+ebx*4]      ;get opcode type and operands
    mov [.OpcodeInfo],eax           ;store info about opcode type

	test byte [.Options],.OptionsAddress
	jz .NoAddress
	mov eax,esi                     ;copy program counter address
    mov ecx,16                      ;sixteen bits
    call MakeHexNum
    mov dword [edi],"    "          ;separate address
    add edi,byte 2
.NoAddress:

    ;(edx=opcode)
	test byte [.Options],.OptionsRegs
	jz .NoRegs
    mov eax,"Y:"<<16
    mov bl,[SpcEmu.Y]               ;and Y
    call .RegisterValue
    mov eax,"A:"<<16                ;set name of register
    mov bl,[SpcEmu.A]               ;get value of SPC register A
    call .RegisterValue             ;print value of register with trailing space
    mov eax,"X:"<<16
    mov bl,[SpcEmu.X]               ;same with X
    call .RegisterValue
    mov eax,"S:"<<16
    mov bl,[SpcEmu.Sp]              ;stack pointer
    call .RegisterValue
    mov eax,"O:"<<16
    mov bl,dl                       ;opcode value
    call .RegisterValue
    mov eax,"M:"<<16
    mov ebx,[SpcEmu.Dp]
    mov bl,byte [.Opcode+1]         ;get first operand
    mov bl,[SpcRam+ebx]             ;get value at dp mem
    call .RegisterValue
    ;mov eax,"C:"<<16
    ;mov bl,[SpcEmu.Timer0]
    ;call .RegisterValue
    stosb                           ;separate with a double space
.NoRegs:

    ;(edx=opcode)
	test byte [.Options],.OptionsFlags
	jz .NoFlags
	mov esi,SpcFlagNames
	mov cl,8
	mov ch,[SpcEmu.Flags]
.NextFlag:
	lodsw                           ;get flag letter and mask
	test ch,ah
	jz .FlagNotSet
	and al,~32                      ;capitalize flag letter if set
.FlagNotSet:
	stosb
	dec cl
	jnz .NextFlag
	mov ax,"  "
	stosw
.NoFlags:

    ;(edx=opcode)
	test byte [.Options],.OptionsOps
    jz near .NoOps
    movzx ecx,byte [SpcEmu.Sp]      ;stack pointer
    not cl
    mov bl,cl
    sub cl,[.SpBase]
    jc .IndentTooFarLeft
    cmp cl,.MaxOpcIndent
    jbe .IndentOps
    mov cl,.MaxOpcIndent
    sub bl,cl
    mov [.SpBase],bl
    jmp .IndentOps
.IndentTooFarLeft:
    mov [.SpBase],bl
    xor cl,cl
.IndentOps:
    mov al," "
    rep stosb
    movzx ebx,byte [.OpcodeInfo]    ;get opcode type
    lea esi,[SpcOpcNames+1+ebx*8]   ;point to mnemonic name for type
    movzx ecx,byte [esi-1]          ;get mnemonic length
    rep movsb                       ;copy string name
    movzx ebx,byte [.OpcodeInfo+1]  ;get type of first operand
    cmp bl,Opr.None                 ;quit if none
    je .NoOps
    mov al," "                      ;separate next operand with a space
    mov esi,[SpcOprNames.Table+ebx*4];get string to make
    call .OprString
    movzx ebx,byte [.OpcodeInfo+2]  ;get type of second operand
    cmp bl,Opr.None
    je .NoOps
    mov al,","                      ;separate next operand with a comma
    mov esi,[SpcOprNames.Table+ebx*4];get string to make
    call .OprString
.NoOps:
.Done:
    mov dword [edi],0A0Dh
    ret

;(esi=formatted string, eax=first character 0-0FF, ebx=operand type 0-0FF, edi=buffer)
;Prints a formatted operand string into the buffer
.OprString:
    movzx eax,al                    ;zero out top bits
.OprStringContinue:
	stosb                           ;write precharacter or last literal
.OprStringNext:
	lodsb                           ;get next character in string
	cmp al,SpcOprNames.HighestCode  ;check if literal character or code
    ja .OprStringContinue           ;continue in loop if literal
	jmp [.OprJumpTable+eax*4]
;.OprEnd: see below in RegisterValue
.OprOpcode:
	mov al,dl                       ;copy opcode
	shr al,4                        ;get top four bits
    mov al,[eax+Text.NumberTable]   ;make ASCII character
    jmp short .OprStringContinue
.OprOpcode2:
	mov al,dl                       ;copy opcode
    shr al,5                        ;get top three bits
    add al,'0'                      ;make ASCII character
    jmp short .OprStringContinue
.OprWord:
	mov eax,edx                     ;copy operand
	shl eax,8                       ;adjust word to top of dword
	mov ecx,4                       ;four hex chars
    jmp short .OprNum
.OprByte:
	mov eax,edx                     ;copy operand
	shl eax,16                      ;adjust byte to top of dword
	mov ecx,2                       ;two hex chars
;(eax=value, ecx=nybbles in number)
.OprNum:
    mov ebx,Text.NumberTable
.OprNumNextDigit:
	and eax,0FFFF0000h              ;mask off unused bits
	rol eax,4                       ;bring up next four bits
	xlatb                           ;turn nybble into ASCII character
	stosb
	dec ecx
    jnz .OprNumNextDigit
	xor eax,eax                     ;rezero for pointer use
    jmp short .OprStringNext
.OprMemBit:
	mov eax,edx                     ;copy operand
    and eax,1FFF00h                 ;get address
    mov ecx,16                      ;sixteen bits
    push esi
    shl eax,8                       ;move address to top of dword
    call MakeHexNum.NoAdjust
    pop esi
	mov eax,edx                     ;copy operand
    and eax,0E00000h                ;get selected bit
    shr eax,13                      ;put into ah
    add eax,'.0'                    ;make ASCII character
    stosw
	xor eax,eax                     ;rezero for pointer use
    jmp short .OprStringNext
.OprJumpRel:
    movzx ecx,byte [.OpcodeInfo+3]  ;get opcode byte length
    movsx eax,byte [.Opcode+ecx-1]  ;get relative distance (-128 - +127)
    add eax,ecx                     ;add opcode length to offset
    add eax,[SpcEmu.Pc]             ;add program counter offset
	shl eax,16                      ;adjust word to top of dword
	mov ecx,4                       ;two hex chars
    jmp short .OprNum
.OprDpO:
    mov al,dh                       ;copy first operand
.OprDpOMututal:
    test byte [SpcEmu.Flags],SpcFlag.P
    setnz ah
    shl eax,20                      ;adjust 12 bits to top
    mov ecx,3
    jmp short .OprNum
.OprDpO2:
    mov al,[.Opcode+2]              ;copy second operand
    jmp short .OprDpOMututal

; Prints a register name and byte value.
; (eax=register name, bl=byte value, edi=buffer)
.RegisterValue:
    movzx ecx,bl
    and ebx,15                      ;make ptr of lower nybble
    shr ecx,4                       ;make ptr of upper nybble
    mov ah,[ebx+Text.NumberTable]   ;get ASCII character
    mov al,[ecx+Text.NumberTable]   ;get ASCII character
    rol eax,16                      ;make register name come first
    stosd
    mov al," "
    stosb
.OprEnd:
    ret

;==============================
section .data
.Options:       db -1   ;which parts to display
.OptionsAddress equ 1
.OptionsFlags   equ 2
.OptionsRegs    equ 4
.OptionsOps     equ 8
.SpBase         db 0    ;used for indentation
 align 4
.OprJumpTable:  ;! must be in the same order as the values in SpcOprNames
                dd .OprEnd,.OprByte,.OprWord,.OprOpcode,.OprMemBit,.OprJumpRel,.OprOpcode2,.OprDpO,.OprDpO2
section .bss
 alignb 4
.Opcode:        resd 1  ;holds current opcode byte value and operand values
.OpcodeInfo:    resd 1  ;holds info about opcode like type, size, operands
.Buffer:        resb 256
section .text

;==============================
; Prints a hex value of varying digit length, from nybble to word.
; (upper eax=byte/word/dword value, ecx=bits, edi=buffer) (edi=byte after last character output)
; (saves edx)
MakeHexNum:
    ror eax,cl              ;adjust nybble/byte/word to top of dword
;(eax=value at top of dword)
.NoAdjust:
    cld
.NextDigit:
    rol eax,4               ;bring up next four bits
    mov ebx,eax
    and ebx,15              ;make pointer
    lea esi,[Text.NumberTable+ebx]
    movsb
    sub ecx,byte 4
    ja .NextDigit
	ret
