;Spc2Midi - Note output through FM synthesis

Sb:
.Timer1Count    equ 2   ;incremented every 80 microseconds
.Timer2Count    equ 3   ;incremented every 320 microseconds
.TimerControl   equ 4
.TimerReset     equ 128 ;clears timer expired flags in port 388h
.Timer1Mask     equ 64
.Timer2Mask     equ 32
.Timer1Start    equ 1
.Timer2Start    equ 2

section .text

;====================
; Turns a channel on (using the current patch for that channel)
;
;(esi=ptr in note buffer, eax=voice)
SoundNoteOn:
    bts dword [Sound.ChannelsOn],eax    ;flag that voice was turned on recently
    btr dword [Sound.ChannelsOff],eax   ;clear that voice needs to be turned off
    bts dword [Sound.ChannelsUsed],eax  ;flag voice is in use
    mov edi,eax                 ;make copy of voice
    jnc .AlreadyOff
    add al,0B0h                 ;DSP register, low byte of pitch
    ;xor ah,ah                   ;turn voice off
    call WriteSbReg
.AlreadyOff:

    mov dl,[NoteBuffer+esi+5]   ;get nearest half-semitone
    movzx ebx,byte [NoteBuffer+esi+6] ;get sample number
    mov ebx,[DspSampleTable.AtrsIdx+ebx*4]
    sub edx,208                 ;reduce 4096 playrate value from 208 to 0
    shl ebx,SampleAtrs_Shift
    add dl,[SampleAtrs.HalfSemitone+ebx]
    and edx,11111110b           ;get MIDI note number (*2)
    lea eax,[edi+0A0h]          ;DSP register, low byte of pitch
    mov ebx,[SbFreqTable+edx]
    mov ah,bl
    call WriteSbReg
    lea eax,[edi+20B0h]         ;DSP register, high byte of pitch (20h turns voice on)
    or ah,bh
    jmp WriteSbReg
;.End:
;    ret

;====================
;(esi=ptr in note buffer, eax=voice)
SoundNoteOff:
    bt dword [Sound.ChannelsOn],eax     ;was voice recently turned on
    jc .Later                           ;if so, turn if off the next call
.Now:
    btr dword [Sound.ChannelsUsed],eax  ;clear voice as being used
    add al,0B0h                         ;DSP register, low byte of pitch
    ;xor ah,ah                          ;turn voice off
    jmp WriteSbReg
.Later:
    bts dword [Sound.ChannelsOff],eax   ;simply flag voice as needing to be turned off
    ret

;==============================
GetSbAddress:
    ;get DOS environment string (db "BLASTER",0)
    ;return error if no environment string
    ;get parts  A,D,I,H
    ;return error if missing i/o address
    jmp .SkipNext
.Next:
    push eax
    inc esi
    call GetHexNum
    ;jz .Error
    pop ebx
    and bl,~32                  ;make uppercase
    cmp bl,'A'
    je .Adr
    cmp bl,'D'
    je .Dma
    cmp bl,'H'
    je .Hdma
    cmp bl,'I'
    jne .Skip
.Int:
    ;mov [],al
    jmp short .Skip
.Adr:
    ;mov [],ax
    jmp short .Skip
.Dma:
    ;mov [],al
    jmp short .Skip
.Hdma:
    ;mov [],al
    ;jmp short .Skip
.Skip:
    mov al,[esi]
    inc esi
    cmp al,' '
    jne .Error
.SkipNext:
    mov al,[esi]
    cmp al,'A'
    jae .Next
    inc esi
    test al,al
    jnz .SkipNext
    clc
    ret
.Error:
    stc
    ret

GetHexNum:
    ret
;==============================
; Enables different waveforms, disables rhythm mode to make use of all 9
; channels, sets default patch sounds.
SbFmInit:
    test byte [PlayOptions],PlayOptions.Sounds
    jz near .End
    mov esi,Text.InitializingSound
    call AddMessage

    xor ebx,ebx                 ;turn off all registers (may be unnecessary)
.NextRegZero:
    mov eax,ebx
    call WriteSbReg             ;destroys eax,ecx,edx
    inc bl                      ;next voice
    jnz .NextRegZero

    mov eax,2001h               ;enable different waveforms
    call WriteSbReg
    mov eax,00BDh               ;disable rhythm mode
    call WriteSbReg
    mov edi,9-1                 ;start with last voice
    mov esi,SbFm.PatchDefault
.NextVoicePatch:
    mov eax,edi
    call SetVoicePatch
    dec edi
    jns .NextVoicePatch

;    mov dword [Sound.ChannelsOff],-1
;    mov ebx,00B0h               ;turn off voices
;.NextVoiceOff:
;    mov eax,ebx
;    call WriteSbReg             ;destroys eax,ecx,edx
;    inc ebx                     ;next voice
;    cmp bl,0B8h
;    jbe .NextVoiceOff

    cld
    mov edi,Sound.UsedInstruments
    xor eax,eax                 ;!!! change to 0FFh later
    stosd
    stosd
    stosb
.End:
    clc                         ;no error
    ret

;==============================
MpuInit:
    ;ret

;==============================
PlayNotes:
    ;check for notes that are different between this call and the last
    ;sound any new notes, release old notes
    ;read in notebuffer from last point read up to cycle time
    ;if note press
    ;  if voice's instrument is different from the one last used, then set patch
    ;  set volume accordingly
    ;  set pitch
    ;  turn on note
    ;if note release
    ;  turn off note
    ;if pitch change
    ;  set note's new pitch
    ;  turn on note
    ;
    ;There a voices which:
    ;  1. are currently in use              ChannelsUsed
    ;  2. were just turned on in this call  ChannelsOn
    ;  3. need to be turned off next call   ChannelsOff

;Turn off any channels which were not turned off last interrupt call
    mov esi,[Sound.ChannelsOff]
    mov dword [Sound.ChannelsOn],0
    test esi,esi
    jz .NoVoicesOff             ;no voices to turn off
    mov ebx,8
.NextVoiceOff:
    bt esi,ebx
    jnc .NoVoiceOff
    mov eax,ebx
    btr dword [Sound.ChannelsUsed],ebx
    add al,0B0h                 ;DSP register, low byte of pitch, voice off
    call WriteSbReg
.NoVoiceOff:
    dec ebx
    jns .NextVoiceOff
    mov dword [Sound.ChannelsOff],0
.NoVoicesOff:

    mov esi,[NoteBufferPlayed]
    cmp [NoteBufferEnd],esi
    je near .End
.NextNote:
    mov eax,[NoteBuffer+esi]    ;get cycle time when note was recorded
    cmp [CycleCount],eax
    jb near .End
    movzx eax,byte [NoteBuffer+esi+4]
.KeyOn:
    test al,1
    jnz .KeyOff
    shr eax,4                   ;select voice
    movzx ebx,byte [NoteBuffer+esi+6] ;get sample number
    bt dword [ActiveChannels],eax
    jnc .AfterKeyChange
    mov edi,[DspSampleTable.AtrsIdx+ebx*4]
    shl edi,SampleAtrs_Shift
    mov bl,[SampleAtrs.Instrument+edi]
    cmp [Sound.UsedInstruments+eax],bl
    je .SameInstrument
    mov [Sound.UsedInstruments+eax],bl
    push eax                    ;save voice
    push esi                    ;save pointer in note buffer
    shl ebx,4                   ;*16 get instrument information
    lea esi,[SbFm.PatchData+ebx];get ptr to instrument information in patch table
    call SetVoicePatch
    pop esi
    pop eax
.SameInstrument:
    call SoundNoteOn
    jmp short .AfterKeyChange
    ;jmp short .DebugPause
.KeyOff:
    test al,2
    jnz .KeyPitchChange
    shr eax,4                   ;select voice
    bt dword [ActiveChannels],eax
    jnc .AfterKeyChange
    call SoundNoteOff
    ;jmp short .AfterKeyChange
.KeyPitchChange:
    ;call something...
.AfterKeyChange:
    add esi,byte 8
    and esi,(NoteBufferSize-1)^7
    cmp [NoteBufferEnd],esi
    jne near .NextNote
.End:
    mov [NoteBufferPlayed],esi
    ret

;Little debugging to stop on each note
;.DebugPause:
;    and byte [PlayOptions],~PlayOptions.Timed
;    add esi,byte 8
;    and esi,(NoteBufferSize-1)^7
;    mov [NoteBufferPlayed],esi
;    ret


;==============================
; (esi=patch info, al=voice) - preserves esi, edi
; Voices are zero based.
SetVoicePatch:
    movzx eax,al
    xor ebx,ebx                 ;start with first attribute
    push dword [SbFm.VoiceTable+eax] ;save voice operator offset
    jmp short .FirstReg
.NextReg:
    add al,[esp]                ;add voice operator offset
    mov ah,[esi+ebx]
    inc ebx                     ;next voice attribute
    call WriteSbReg
.FirstReg:
    mov al,[SbFm.PatchRegs+ebx]
    test al,al
    jnz .NextReg
.End:
    pop eax                     ;release saved operator offset
    ret

;==============================
;
; Quote from "Programming the AdLib/Sound Blaster" by Jeffrey S. Lee (1992)
;
; After writing to the register port, you must wait twelve cycles before
; sending the data; after writing the data, eighty-four cycles must elapse
; before any other sound card operation may be performed.
;
; The AdLib manual gives the wait times in microseconds: three point three
; (3.3) microseconds for the address, and twenty-three (23) microseconds
; for the data.
;
; The most accurate method of producing the delay is to read the register
; port six times after writing to the register port, and read the register
; port thirty-five times after writing to the data port.
;
; (al=sb reg, ah=value) - preserves ebx,esi,edi
;
WriteSbReg:
    mov edx,[Sound.FmPort]
    out dx,al
    mov ecx,2
    shl eax,8       ;preserve register value in upper dword
    call TimedWait  ;delays 15-30ms
    shr eax,16      ;restore register value from upper dword
    inc edx
    out dx,al
    mov ecx,3       ;delays 30-45ms
    ;jmp TimedWait

;====================
; Quote from "CPU Independant Timer"
;
; "the perfect anwser.  It seems that there is a timer connected to port 61h
;  at bit 4.  This bit toggles every 15.085 u sec (micro sec)."
;
; (ecx=number of 15.085 microsecond loops) () - Only changes ax and ecx
;
TimedWait:
    in al,61h
    and al,10h          ;mask bit 4
    mov ah,al
.CheckAgain:
    in al,61h
    and al,10h
    cmp al,ah
    jz .CheckAgain      ;no change so loop back
    mov ah,al
    dec ecx
    jnz .CheckAgain
    ret

;==============================
; Rather than just abruptly turn off all the notes, it first sets the release
; rates so that they fade out.
SilenceNotes:
    ret
    test byte [cs:PlayOptions],PlayOptions.Sounds
    jz .End

    mov ebx,00B0h               ;turn off voices
.NextVoiceOff:
    mov eax,ebx
    call WriteSbReg             ;destroys eax,ecx,edx
    inc ebx                     ;next voice
    cmp bl,0B8h
    jbe .NextVoiceOff

    mov ebx,0080h
.Next:
    mov eax,ebx
    mov ah,054h                 ;sustain/release
    call WriteSbReg             ;destroys eax,ecx,edx
    mov ah,075h                 ;attack/decay
    call WriteSbReg             ;destroys eax,ecx,edx
    inc ebx                     ;next voice
    cmp bl,060h+31
    jbe .Next
.End:
    ret

;========================================
Sound:
section .data
align 4
.FmPort:        dw 388h
.MidiPort:      dw 330h
.ChannelsOn:    dd 0    ;channels recently turned on
.ChannelsOff:   dd -1   ;channels to turn off soon
.ChannelsUsed:  dd 0    ;channels currently used by voices
.Dma:
.Interrupt:
.Supported:     db 1    ;flags whether or not sound card is present

section .bss
.UsedInstruments:resb 9  ;holds the last used instrument for that voice

section .data
SbFm:
.PatchRegs:     db 20h,23h      ;multiplier
                db 40h,43h      ;level
                db 60h,63h      ;attack/decay
                db 80h,83h      ;sustain/release
                db 0E0h,0E3h    ;waveform
                db 0C0h         ;feedback
                db 0            ;null (end)
.VoiceTable:    db 0,1,2,8,9,10,16,17,18,24,25,26
align 4
.PatchData:     incbin "patches.dat"
.PatchDefault:  db   2h,  1h;2
                db  10h,  0h
                db  99h,074h;f4
                db  88h,0FFh
                db   0h,  0h
                db   0h
