; Spc2Midi - DSP to MIDI converter
; PeekinSot (c)2000
;
; By Dwayne Robinson (FDwR@hotmail.com)
; http://members.tripod.com/FDwR/snes.htm
; Assembly compiled with NASM and WDOSX
;
; Major Thanks to:
;   Ledi for the APU manual
;   _Demo_ for letting me see the SPCPLAY 2.0 source code
;   zsKnight for answering my annoying questions
;   Gau for his SPC document
;   CyberGoth for encouragement
;   Authors of NASM and WDOSX for the great tools to create this program
;
;   And of course, the great SNES music writers (like Koji Kondo!)
;   who gave me a reason to write this
;
; Notes on note conversion:
;   - A 16k table, along with the sample info table, is used to convert 14bit
;     pitches to MIDI notes 0-127. In human hearing, a note that is twelve
;     semitones above another (or one whole octave), is actually twice as
;     high in frequency. If a note played at rate of 4096 (32khz) sounded a
;     middle C, 2048 would be one octave below, and 8192 would be an octave
;     above. The table is used to convert the exponential curve of frequency
;     to a linear value from 0 to 255, with the playrate of 4096 (sample
;     played at actual frequency) being 208. Then, the semitone value from the
;     sample info table is added to it to get the actual MIDI note value.
;     The semitone value from the sample table is the pitch that the BRR wave
;     sample sounds closest to when played at true frequency (4096 or 32khz).
;
;--------------------------------------------------

%define ProgVersion     ".105"
SpcDebugMode            equ 0   ;1=disassembly 2=graphical path 3=dsp writes as pixels
SpcDebugDspOutput       equ 0
SpcDebugIoPrint         equ 1

CyclesPerSecond         equ 2496000
CyclesPerTimerTick      equ 39
CyclesToEmulateAhead    equ CyclesPerSecond/4 ;emulate ahead by a fourth of a second
;CyclesPerSbTick        equ 83200   ;CyclesPerSecond / 30.04808  (320us * 104)
CyclesPerPcInterrupt    equ CyclesPerSecond/30 ;CyclesPerSecond / 18.2065   (1.1931817MHz / 65536)
DspEventBufferSize      equ 16384   ;holds 2048 dsp writes
NoteBufferSize          equ 65536   ;holds 8192 note presses/releases
Opr.None                equ 0
NumStringMaxLen         equ 10

StackReg:
.Di     equ 0
.Si     equ 4
.Bp     equ 8
.Sp     equ 12
.Bx     equ 16
.Dx     equ 20
.Cx     equ 24
.Ax     equ 28

ZstFile:;relative positions of registers within the savestate's spc portion
.Pc     equ 0
.A      equ 4
.X      equ 8
.Y      equ 12
.Flags  equ 16
.NZ     equ 20
.Sp     equ 24

SpcFile:;relative positions of registers
.Pc     equ 0
.A      equ 2
.X      equ 3
.Y      equ 4
.Flags  equ 5
.Sp     equ 6
;0002Eh-0004Dh - SubTitle/Song Name
;0004Eh-0006Dh - Title of Game
;0006Eh-0007Dh - Name of Dumper
;0007Eh-0009Dh - Comments
;0009Eh-000A4h - Date of SPC Dumped in decimal (DD/MM/YYYY)
;000A9h-000ABh - Time in seconds for the spc to play before fading
;000ACh-000AFh - Fade out time in milliseconds
;000B0h-000CFh - Author of Song
;000D0h        - Default Channel Disables (0 = enable, 1 = disable)
;000D1h        - Emulator used to dump .spc file
;                 (0 = UNKNOWN, 1 = ZSNES, 2 = SNES9X)

;============================================================
SECTION .bss
DataSelector:   resw 1  ;necessary for interrupt handler
PspSelector:    resw 1  ;selector to PSP

alignb 4
StartOptions:
.Count:         resd 1  ;number of parameters
.Ptrs:          resd 1  ;array of pointers to each one
.Env:           resd 1  ;command environment strings

alignb 4
SpcRam:         resb 65536
DspRam:         resb 128
SpcEram:        resb 64
SpcRegs:        resb 48
DspRamCopy:     resb 128

ResetSpcRam:    resb 65536
ResetDspRam:    resb 128
ResetSpcEram:   resb 64
ResetSpcRegs:   resb 48

alignb 4
%if SpcDebugMode=3
    DspEventBuffer      equ 0A0000h
%else
    alignb 4
    DspEventBuffer:     resb DspEventBufferSize
%endif
NoteBuffer:             resb NoteBufferSize

PitchTable:             resb 16384  ;for converting an absolute pitch into a half-semitones

; Stores information for all kinds of sound effects used in a game. I would
; be very surprised if a game used more than 512 instruments, or even more
; than 512 sound effects in general. There is always at least one entry
; defined which can not be deleted.
SampleAtrs:
.Length:                resd 1      ;sample length (not bytes)
.LoopLen:               resd 1      ;length of loop portion
.Checksum:              resd 1      ;addition checksum of both sample and loop
.HalfSemitone:          resb 1      ;nearest note
.Instrument:            resb 1      ;MIDI instrument or patch (0-127)
.Volume:                resb 1      ;
.Bank:                  resb 1      ;for MIDI sequencers
.Interpolation:         resb 1      ;simply for possible compatibility with SpcTool
.Name:                  resb 64+$-SampleAtrs
                        resb ($-SampleAtrs)*511
SampleAtrs_Shift        equ 6
SampleAtrsEntry_Size    equ 64

; Stores information about each sound effect currently in SPC memory, pointed
; to by the DSP table. Each sample may have other attributes associated with
; it from the SampleInfo list by an index. The reason why an index into a
; separate table is used rather than storing all the attributes with each
; sample, is because some sample numbers share the same BRR wave sample.
DspSampleTable:
.Used:                  resb 256/8  ;256 bit flags to indicate which samples have been used
.Offset:                resd 256    ;word offsets of sample:loop
.Length:                resd 256    ;sample length (not bytes)
.LoopLen:               resd 256    ;length of loop portion
.Checksum:              resd 256    ;addition checksum of both sample and loop
.AtrsIdx:               resd 256    ;indexes into SampleAtrs for more attributes about this sample
DspSampleTable_Size     equ $-DspSampleTable

SamplesBuffer_Size      equ 232992
SamplesBuffer:          resb SamplesBuffer_Size     ;64k * (8/9) * 4
SamplesBuffered_Size    equ 65536/8
SamplesBuffered:        resb SamplesBuffered_Size   ;flags indicating which parts of memory have been brr decompressed

;============================================================
section .text
BITS 32
GLOBAL _WdosxStart

_WdosxStart:
;==============================
;open and load file
;do
;	print opcode and flags
;	emulate for one cycle
;loop until keypress
;check keypress
;resume emulation or quit
;end
Main:
    mov [StartOptions.Count],esi    ;save the number of parameters,
    mov [StartOptions.Ptrs],edi     ;array of pointers to each one,
    mov [StartOptions.Env],ebp      ;and command environment strings
    mov [PspSelector],es
    mov [DataSelector],ds
    push ds
    pop es

	mov edx,Text.Title
	call WriteString

    ;call CheckStartOptions
    mov esi,[StartOptions.Ptrs] ;get first parameter
    mov esi,[esi+4]             ;get filename
    test esi,esi
    jnz .AtLeastOneParameter
    mov edx,Text.About
    jmp EndProgram.ErrorMessage
  .AtLeastOneParameter:
    mov [SpcFileName],esi
	;call GetAbsoluteSelector
    ;jc near EndProgram.ErrorMessage

    ;get fully qualified filename
    call LoadSpcState
    jc near EndProgram.ErrorMessage  ;end program and print error message
  %if SpcDebugMode=2
	call ShowOpcode.PathInit
  %elif SpcDebugMode=3
    call ShowOpcode.SetVideoMode
  %endif
  %if SpcDebugDspOutput
    mov ah,3Ch                  ;create file
    xor ecx,ecx                 ;no attributes
    mov edx,DspFileName
    int 21h
    mov edx,Text.DspOutputError
    jc EndProgram.ErrorMessage  ;end program and print error message
    mov [DspFileHandle],eax
    mov ebx,eax
    mov ecx,128
    mov edx,DspRam
    mov ah,40h                  ;DOS: Write to file
    int 21h
  %endif

    ;mov ah,1                   ;set cursor type
    ;mov ecx,2000h              ;turn off cursor
    ;int 10h
    ;load configuration file
    ;get filename without extension and add .sps
    ;call LoadSpcSettings
    ;mov esi,Text.LoadingSettings
    ;call AddMessage
    call DspEmuInit             ;copy DSP RAM, decache brr sample table, flag all samples as unused, expand 128 MIDI notes into 16k pitch table
    call BufferAllSamples
    ;call DumpSoundBuffer;.InfoOnly
    call SbFmInit               ;enable waveform select, disable rhythm mode, turn off all voices

    mov edx,Text.Ready
    call WriteString
    xor eax,eax
    int 16h
    cmp al,27
    je .End

%if SpcDebugMode=0
    call GuiInit
    call SpcViewWndSetMode
%endif
    call SetupTimers            ;hook timer interrupt

    call MainLoop

    call SetupTimers.RestoreOldHandler
%if SpcDebugMode=0
    call GuiDeinit
    call SilenceNotes
%endif

.End:
  %if (SpcDebugMode=2) | (SpcDebugMode=3)
    mov eax,3       ;restore text mode
	int 10h
  %endif
  %if SpcDebugDspOutput
    mov ah,3Eh              ;close file
    mov ebx,[DspFileHandle]
    int 21h
  %endif

    ;save configuration file

;==============================
EndProgram:
    mov eax,4C00h
	int 21h
; (edx=message text ptr)
.ErrorMessage:
	call WriteString
    mov eax,4C01h
	int 21h

;============================================================
MainLoop:
;do
;   if CyclesEmulated < CyclesToEmulate then
;       emulate for portion of cycles
;       update window if changed
;       if recording midi then record to midi file
;       if recording dsp writes then record to dsp output file
;   endif
;   check keypresses and act on them
;loop
    call ShowActiveChannels
.NextEmuRun:
    cmp dword [CyclesToEmulate],0   ;is emulation too fast or too slow?
    jle .EmulationTooFast           ;emulation is too fast so slow down some
  %if SpcDebugMode=1
    mov eax,2000
  %else
    mov eax,41600
  %endif
    sub [CyclesToEmulate],eax
	call SpcEmu.Continue
  %if SpcDebugDspOutput
    call OutputDspWrites
  %endif
    ;mov eax,[TestCounter];[CycleCount]
    ;mov ecx,32
    ;mov edi,CounterText
    ;call MakeHexNum
    ;mov ah,9
    ;mov edx,CounterText
    ;int 21h
    ;check for cycle counter wrap by checking sign
    ;test [PlayOptions],PlayOptions.MidiLog
    ;jz .NoMidiOutput
    ;call OutputMidiEvents
;.NoMidiOutput:
    ;test [PlayOptions],PlayOptions.DspLog
    ;jz .NoDspOutput
    ;call OutputDspWrites
;.NoDspOutput:
.EmulationTooFast:
    ;call PlayNotes
    mov eax,[CycleCount]
    cmp [CycleCountPrev],eax
    je .NoUpdate
    mov [CycleCountPrev],eax
    call NoteBufferSeekNext
%if SpcDebugMode=0
    call dword [SpcViewWnd.Handler]
    call dword [SpcViewWnd.ScreenCopier] ;this should later be moved to the GUI redraw
%endif
.NoUpdate:
	mov ah,1
	int 16h
	jz .NextEmuRun

    xor eax,eax
    int 16h
    cmp al,' '
    jne .NotPauseResume
    xor byte [PlayOptions],PlayOptions.Timed
    jmp .NextEmuRun
.NotPauseResume:
    cmp al,8
    jne .NotRestart
    push dword [PlayOptions]
    and byte [PlayOptions],~(PlayOptions.Sounds|PlayOptions.Timed)  ;do not want an interrupt to occur right here
    mov dword [NoteBufferPos],0
    mov dword [NoteBufferPlayed],0
    mov dword [CycleCount],0
    mov byte [Sound.ChannelsOff],255
    call SpcViewWndSetMode.ClearTable
    pop dword [PlayOptions]
    jmp .NextEmuRun
.NotRestart:
    cmp al,'9'
    ja .NotChannel
    je .TurnOnAll
    cmp al,'0'
    jb .NotChannel
    je .TurnOffAll
    movzx eax,al
    sub eax,byte '1'
    btc [ActiveChannels],eax
    jnc .TurnedOn
    bts dword [Sound.ChannelsOff],eax
.TurnedOn:
    call ShowActiveChannels
    jmp .NextEmuRun
.TurnOffAll:
    mov byte [ActiveChannels],0
    mov byte [Sound.ChannelsOff],255
    call ShowActiveChannels
    jmp .NextEmuRun
.TurnOnAll:
    mov byte [ActiveChannels],255
    call ShowActiveChannels
    jmp .NextEmuRun
.NotChannel:
    cmp ax,'H'<<8
    mov bl,1
    je .ChangeInstrument
    cmp ax,'P'<<8
    jne .NotInstrument
    neg bl
.ChangeInstrument:
    mov al,[SampleAtrs.Instrument]
    add al,bl
    and al,127
    mov [SampleAtrs.Instrument],al
    jmp .NextEmuRun
.NotInstrument:
    cmp ax,'M'<<8
    mov ebx,CyclesPerSecond/300
    je .ChangeSpeed
    cmp ax,'K'<<8
    jne .NotChangeSpeed
    neg ebx
.ChangeSpeed:
    add [CyclesPerInterrupt],ebx
    jns near .NextEmuRun
    mov dword [CyclesPerInterrupt],0
    jmp .NextEmuRun
.NotChangeSpeed:
.Done:
    ret

    ;<Keys while playing>
    ;Insert = Set mark
    ;Delete = Delete mark
    ;Home   = Previous mark
    ;End    = Next mark
    ;PgDown = Page backward
    ;PgUp   = Page forward
    ;BkSpc  = Restart
    ;Space  = Pause/Resume
    ;0-9             = Toggle channels/Change value
    ;Up/Down         = Select option
    ;Left/Right      = Small increment
    ;Ctrl+Left/Right = Large increment
    ;* / + -

;==============================
; This is called upon interrupt 8 from the pc timer (18.2hz).
; There are 3 possibilities. Emulation is far ahead of actual time so a that
; an emulation wait is required, actual time is ahead of emulated time so
; that emulation must be given a chance to catch up, or emulation is just
; ahead of actual time.
;
; //only if emulation is up to or ahead of realtime, not lagging behind
; if (CyclesEmulated - CycleCount <= 0) then
;     CyclesToEmulate = CyclesToEmulateAhead
; elif (CyclesEmulated - CycleCount <= CyclesToEmulateAhead) then
;     CycleCount += CyclesPerInterrupt
;     CyclesToEmulate = CyclesToEmulateAhead
; else
;     CycleCount += CyclesPerInterrupt
; endif
;
;  (OR)
;
; if (CyclesEmulated - CycleCount <= 0) then
;     CyclesToEmulate = CyclesToEmulateAhead      'emulation is behind
; else
;     CycleCount += CyclesPerInterrupt            'emulation is okay or fast
;     if (CyclesEmulated - CycleCount <= CyclesToEmulateAhead) then
;         CyclesToEmulate = CyclesToEmulateAhead  'emulation is okay
;     endif                                       'else it is fast
; endif
; if (playing sounds) then
;     save all registers
;     read note buffer up to current time and play sounds
;     restore all registers
; endif
;
TimerHandler:
    push ds
    push es
    push eax
    mov ax,[cs:DataSelector]
    mov ds,ax
    mov es,ax

    ;increment GUI time counter

    test byte [PlayOptions],PlayOptions.Timed
    jz .NoTimerAdvance
    push ebx
    mov eax,[CyclesEmulated]
    mov ebx,[CyclesPerInterrupt]
    sub eax,[CycleCount]        ;get how far emulation is ahead of actual time
    jbe .EmulationIsSlow
    add [CycleCount],ebx        ;advance actual time
    cmp eax,CyclesToEmulateAhead
    ja .EmulationIsFast
.EmulationIsSlow:
    mov dword [CyclesToEmulate],CyclesToEmulateAhead  ;advance emulation time
.EmulationIsFast:
    test byte [PlayOptions],PlayOptions.Sounds
    jz .NoSounds
    pusha
    call PlayNotes
    popa
.NoSounds:
    pop ebx

.NoTimerAdvance:
    add dword [InterruptAccum],(1193182/30)<<16 ;18.2 / 30 ratio
    jnc .NoChain                ;take care of int ourselves
    pop eax
    pop es
    pop ds
    jmp far [cs:SetupTimers.OldHandler]
.NoChain:
    mov al,20h                  ;acknowledge interrupt
    out 20h,al
    ;sti                         ;?is this necessary
    pop eax
    pop es
    pop ds
    iret

%if 0
    push eax
    mov al,20h
    out 20h,al      ;End of interrupt
    sti
    iretd
.Chain:
    pop eax
    pop edx
    jmp far [cs:SetupTimers.OldHandler]
%endif

;============================================================
; Reads in a savestate or spc, registers, flags, memory...
; Ends if:
;   The filetype has an unknown extension
;   The file does not exist
;   It is corrupt
;   If the savestate was not saved with sound data
LoadSpcState:
    mov esi,[SpcFileName]
    call GetFileExtension
    mov ebx,'SPC'
    call CheckFileExtensionMatch
    je near .Spc
    mov ebx,'ZST'
    call CheckFileExtensionMatch
    je .Zst
    mov ebx,'ZMV'
    call CheckFileExtensionMatch
    je .Zst
.FileTypeError:
    mov edx,Text.FileTypeBad                ;don't know how to load it
    stc
	ret

.Zst:
    mov esi,ZstFilePositions
    call .ReadParts
    jc near .FileOpenError
    test byte [.SpcSaved],1                 ;check that sound was enabled
    jz near .ReadPartError                  ;when the savestate was saved

	;Convert flags
    movzx eax,byte [SpcRegs+ZstFile.Flags]  ;get flags
    mov al,[SpcEmu.FlagTable+eax]           ;remap flags
    mov bl,byte [SpcRegs+ZstFile.NZ]        ;get SpcNZ
	and al,~(SpcFlag.N|SpcFlag.Z)           ;mask out N and Z
	test bl,bl
    jnz .ZeroFlagNotSet
    or al,SpcFlag.Z                         ;set zero flag if NZ is zero
.ZeroFlagNotSet:
    and bl,SpcFlag.N                        ;isolate N
	or al,bl                                ;combine N with flags

    ;Copy registers
    mov dx,[SpcRegs+ZstFile.Pc]
    mov bl,[SpcRegs+ZstFile.A]
    mov cl,[SpcRegs+ZstFile.X]
    mov bh,[SpcRegs+ZstFile.Y]
    mov ch,[SpcRegs+ZstFile.Sp]
    jmp .Init

.Spc:
    mov esi,SpcFilePositions
    call .ReadParts
    jc near .FileOpenError

    ;Copy registers
    movzx eax,byte [SpcRegs+SpcFile.Flags]  ;get flags
    mov dx,[SpcRegs+SpcFile.Pc]
    mov bl,[SpcRegs+SpcFile.A]
    mov cl,[SpcRegs+SpcFile.X]
    mov bh,[SpcRegs+SpcFile.Y]
    mov ch,[SpcRegs+SpcFile.Sp]
    mov al,[SpcEmu.FlagTable+eax]           ;remap flags
    ;jmp short .Init

; Either spc or zst, both share the same code below.
; (al=flags, bx=AY, cl=X, ch=Sp, dx=Pc)
.Init:
    mov [SpcEmu.Pc],dx
    mov [SpcEmu.AY],bx
    mov [SpcEmu.X],cl
    mov [SpcEmu.Sp],ch
	mov [SpcEmu.Flags],al
	test al,SpcFlag.P                       ;check direct page
	setnz [SpcEmu.Dp+1]                     ;set pointer accordingly
    mov eax,[SpcRam+SpcReg.Timer0]          ;get all three timer values
    and eax,0FFFFFFh                        ;zero top byte
    or eax,0E0000000h                       ;set timer 0,1 8x multiple to max (7*32)
    mov [SpcEmu.Timer0],eax                 ;set timers 0,1,2
    mov byte [DspRam+DspReg.Endx],255       ;set flag that all samples have ended - temporary hack
    mov esi,SpcEram
    mov edi,SpcRam+0FFC0h
    mov ecx,64/4
    cld
    rep movsd
    ;clc
    ret

; Opens the file, reads in the main parts of spc or savestate, then closes it.
; (esi=part table)
.ReadParts:
    mov eax,3D40h           ;open for read with full sharing
    mov edx,[SpcFileName]
    int 21h
    mov edx,Text.FileOpenError
    jc .FileOpenError
    push esi
    push eax                    ;save returned file handle
    mov esi,Text.LoadingFile
    call AddMessage
    pop ebx
    pop esi
    jmp short .FirstPart
.NextPart:
    mov edx,[esi+4]             ;get file position source
    shld ecx,edx,16             ;upper word into cx
    mov eax,4200h               ;get/set file position
    int 21h
    mov edx,edi                 ;get destination to read into memory
    mov ecx,[esi+8]             ;get length of piece
    add esi,byte 12
    mov ah,3Fh                  ;DOS: Read file
    int 21h
    cmp eax,ecx                 ;check that the full amount was read in
    jb .ReadPartError           ;end if amount read was less than it should be
.FirstPart:
    mov edi,[esi]               ;get next memory destination
    test edi,edi
    jnz .NextPart
.CloseFile:
    mov ah,3Eh
    int 21h
.FileOpenError:
    ret

.ReadPartError:
    mov edx,Text.SpcReadError
    stc
    ret

section .bss
.SpcSaved:  resb 1  ;flags whether or not savestate contains SPC info
;.Handle:   resd 1
section .text

;============================================================
; Saves only header information; registers, flags, and memory; or both.
; Handy for cutting down on the size of savestates.
; Ends if:
;   The file does not already exist using partial write mode
;   The file can be written to
%if 0
SaveSpcState:
    ;open existing or create new one
    mov ebx,3D41h               ;open for write with full sharing
    mov edx,[SpcFileName]
    int 21h
    mov edx,Text.FileOpenError
    jc .FileOpenError
    push eax                    ;save returned file handle
    mov esi,Text.SavingFile
    call AddMessage
    pop ebx
.CloseFile:
    mov ah,3Eh
    int 21h
    ret

; Writes out the main parts of the spc.
; (esi=part table)
;.WriteParts:
;    jmp short .FirstPart
.NextPart:
    mov edx,[esi+4]             ;get file position destination
    shld ecx,edx,16             ;upper word into cx
    mov eax,4200h               ;get/set file position
    int 21h
    mov edx,edi                 ;get source to read from memory
    mov ecx,[esi+8]             ;get length of piece
    add esi,byte 12
    mov ah,40h                  ;DOS: Write file
    int 21h
    cmp eax,ecx                 ;check that the full amount was read in
    jb .WritePartError          ;end if amount written was less than it should be
.WriteParts:
.FirstPart:
    mov edi,[esi]               ;get next memory source
    test edi,edi
    jnz .NextPart
.FileOpenError:
    ret

.WritePartError:
    mov edx,Text.SpcReadError
    stc
    ret
%endif

;==============================
; Returns extension without period and top byte null.
;
; (esi=start of filename) (eax=extension, zf=no extension)
GetFileExtension:
    mov edi,esi     ;copy source for string length search
    mov ecx,256     ;make a copy of max length for later and set al to null
    xor eax,eax     ;look for null
    cld
    repne scasb     ;edi will point to one after end of string
    dec edi         ;skip backwards over null
.Next:
    dec edi         ;step back one character
    cmp edi,esi     ;check that it's still in the string
    jbe .End
    mov al,[edi]    ;get character
    cmp al,'.'      ;if period, then it found the extension
    je .Done
    cmp al,'\'      ;if backslash is encountered, then there was no extension
    jne .Next
.End:
    xor eax,eax     ;clear extension and set zero flag
    ret

.Done:              ;capitalize character
    mov eax,[edi+1] ;get extension without period prefix
    mov cl,3
  .NextChar:
    cmp al,'a'
    jb .Lowercase
    cmp al,'z'
    ja .Lowercase
    and al,~32      ;capitalize character
  .Lowercase:
    rol eax,8       ;next character
    dec cl
    jns .NextChar
    ;zero flag should be clear here
    ret

;==============================
; Sets zero flag if file extension matches given pattern. Does not change
; either eax or ebx for the purpose of multiple calls with different
; filename extensions.
;
; (eax=extension, ebx=mask) (zf=match or equal)
CheckFileExtensionMatch:
    cmp eax,ebx         ;compare extension to entire mask
    je .Done            ;match, end
    cmp ax,bx           ;compare first two digits
    jne .Done           ;mismatch, end
    shld ecx,eax,16     ;get final extension digit to see if it's a number
    cmp cl,'0'          ;if less than zero or greater than nine
    jb .Done            ;then mismatch
    cmp cl,'9'
    ja .Done
    cmp eax,eax         ;set zero flag for je
.Done:
    ret

;============================================================
; Saves older timer interrupt handler and sets new one.
;
SetupTimers:
    ;save existing interrupt handler
    mov ax,0204h            ;get protected-mode interrupt
    mov bl,8                ;timer interrupt
    int 31h
    mov [.OldHandler],edx   ;save offset
    mov [.OldHandler+4],cx  ;save selector

    ;test byte [Sound.Supported],-1  ;check if there is a sound card
    ;jz .SetHandler                  ;if none, skip down
    ; The SB timer2 oddly increments up (every 320us) until a byte overflow
    ; rather than decrementing down to zero, so the value we send it is
    ; [256 - x]. [104 * .00032 = .03328] (roughly 1/30th of a second)
    ; Every ten minutes, it may be a second ahead.
    ;mov eax,(250<<8) | Sb.Timer2Count;152
    ;call WriteSbReg
    ;mov eax,(Sb.TimerReset<<8) | Sb.TimerControl
    ;call WriteSbReg             ;reset timer expired flags
    ;mov edx,SbTimerHandler      ;sound card timer handler
    ;call .SetHandler
    ;mov eax,((Sb.Timer1Mask | Sb.Timer2Start)<<8) | Sb.TimerControl
    ;call WriteSbReg             ;activate timer 2 only

    ;test byte [Sound.Supported],-1  ;check if there is a sound card
    ;jz .SetHandler                  ;if none, skip down
    ;mov eax,(0<<8) | Sb.TimerControl
    ;call WriteSbReg                 ;stop timer

    ;increase rate to more precise rate of 30hz rather than only 18.2
    mov cx,1193182/30
    call .SetPitRate

    mov edx,TimerHandler
;(edx=handler address)
.SetNewHandler:
    mov bl,8            ;timer interrupt
    mov ax,0205h        ;set protected mode interrupt
    mov cx,cs           ;pass our code selector
    int 31h
    ret

.RestoreOldHandler:
    ;restore old interrupt handler
    mov bl,8            ;timer interrupt
    mov ax,0205h        ;set protected mode interrupt
    mov cx,[SetupTimers.OldHandler+4]   ;get selector
    mov edx,[SetupTimers.OldHandler]    ;get offset
    int 31h

    ;restore default rate of 18.2065hz
    xor cx,cx
    ;call .SetPitRate
    ;ret

.SetPitRate:
    cli                 ;don't want any interrupt to mess this up below
    mov al,00110100b    ;(00:11:010:0) set counter 0, lsb/msb, mode 2, binary
    out 43h,al
    in al,61h           ;i/o delay
    mov al,cl
    out 40h,al          ;write counter low byte
    in al,61h           ;i/o delay
    mov al,ch
    out 40h,al          ;write counter high byte
    sti
    ret

section .bss
alignb 4
.OldHandler:    resd 1
                resw 1
section .text

;============================================================
;==============================
; Writes an ASCIIZ string to screen, not terminated by stupid "$"
; (edx=ptr to string) (esi%)
WriteString:
    mov ecx,1024            ;maximum length of characters
	mov edi,edx             ;copy source for string length search
	mov eax,ecx             ;make a copy of max length for later and set al to null
    ;xor al,al              ;;look for null
    cld                     ;as always, look forward
    repne scasb             ;search for the end
    ;neg ecx                ;;get negative count
    ;dec ecx                ;;minus the null at the end
	not ecx                 ;negate count and subtract null at the end
	add ecx,eax             ;get length
; (edx=ptr to string, ecx=count)
.OfLength:
    mov ebx,1               ;console output
    mov ah,40h              ;DOS: Write to file
	int 21h
	ret

;==============================
; (esi=source) (ecx=length) (esi)
GetStringLength:
    mov eax,1024            ;maximum length of characters
; (al=character to search for, eax=maximum length of characters to search)
.UntilChar:
    mov ecx,eax             ;make a copy of max length for later
    mov edi,esi             ;copy source for string length search
    cld                     ;as always, look forward
    repne scasb             ;search for the end, until character is found
    ;neg ecx                ;;get negative count
    ;dec ecx                ;;minus the character at the end
    not ecx                 ;negate count and subtract character at the end
	add ecx,eax             ;get length
    ret

;==============================
; (esi=source of text)
AddMessage:
    test byte [Gui.Status],Gui.StatusActive
    jz .DosPrint
    ret
.DosPrint:
    call GetStringLength        ;returns length in ecx
    lea edi,[esi+ecx]
    push dword [edi]
    push edi
    mov word [edi],0A0Dh
    add ecx,byte 2
    mov edx,esi
    call WriteString.OfLength
    pop edi
    pop dword [edi]
    ret

;==============================
; Turns a 32bit number into a decimal (or other) string, writing it to edi.
;
;(eax=number, edi=destination) (ecx=offset in buffer of first significant digit, ebx=radix used)
;(none saved)
NumToString:
    mov ecx,NumStringMaxLen ;default maximum of ten characters, since the largest 32bit number is 4 gigabytes
;(ecx=number of digits)
.OfLength:
	mov ebx,10              ;base of the decimal system
;(ebx=radix)
.OfRadix:                   ;for hexadecimal and binary (even octal)
	xor edx,edx             ;set top 32 bits of quotient to zero
	lea edi,[edi+ecx-1]     ;start from rightmost character in number
.NextChar:
	div ebx                 ;divide number by the decimal base
    mov dl,[Text.NumberTable+edx] ;get ASCII character from table
	;add dl,48              ;make remainder into an ASCII character
	mov [edi],dl            ;output result
	dec edi                 ;move backwards one character
	test eax,eax            ;see if we are done with the number
	jz .FillInBlanks        ;nothing but zeroes left
	xor edx,edx             ;set edx to zero again for next division
	dec ecx                 ;one less character to output
	jnz .NextChar
	ret

.FillInBlanks:
    mov al,[Text.NumberFillChar];fill in with spaces, zeroes, asterisks
	dec ecx                 ;one less than current count
	mov edx,ecx
	std                     ;move backwards
	rep stosb               ;for number of characters remaining
    mov ecx,edx             ;return offset of first digit
	ret

%if 0
;==============================
; ()() zero based selector is stored in AbsoluteSelector
GetAbsoluteSelector:
    xor eax,eax                 ;DPMI function 0, allocate selector
    mov ecx,1                   ;only one selector
	int 31h
	jc .ErrorAllocating	;uh-oh
    mov [AbsoluteSelector],eax  ;store returned selector
    xor edx,edx                 ;set base address to zero
    mov ebx,eax                 ;copy selector
    xor ecx,ecx                 ;put upper 16 bits into ecx (cx:dx)
    mov eax,7                   ;DPMI function 7, set base address
	int 31h
    jc .ErrorSetting            ;could not set base
	xor edx,edx
    mov ecx,(1048576*2)>>16     ;put upper 16 bits into ecx (cx:dx)
    mov eax,8                   ;DPMI function 8, set segment limit
	int 31h
    jc .ErrorSetting            ;could not set limit
	ret

.ErrorSetting:
	mov ebx,[AbsoluteSelector]
	mov eax,1		;DPMI function 1, free selector
	int 31h			;ebx should still contain our selector
	stc
.ErrorAllocating:
	mov edx,Text.ErrorGettingSelector
	ret

align 4
AbsoluteSelector:       dd 0
%endif

;=============================================================================
section .data

align 4
SpcOpcInfo:
incbin "opcinfo.dat"

SpcOpcNames:
db 3,"mov    ",3,"adc    ",3,"sbc    ",3,"cmp    ",3,"and    ",2,"or     "
db 3,"eor    ",3,"inc    ",3,"dec    ",3,"asl    ",3,"lsr    ",3,"rol    "
db 3,"ror    ",3,"xcn    ",4,"movw   ",4,"incw   ",4,"decw   ",4,"addw   "
db 4,"subw   ",4,"cmpw   ",3,"mul    ",3,"div    ",3,"daa    ",3,"das    "
db 3,"bra    ",3,"beq    ",3,"bne    ",3,"bcs    ",3,"bcc    ",3,"bvs    "
db 3,"bvc    ",3,"bmi    ",3,"bpl    ",3,"bbs    ",3,"bbc    ",4,"cbne   "
db 4,"dbnz   ",3,"jmp    ",4,"call   ",5,"pcall  ",5,"tcall  ",3,"brk    "
db 3,"ret    ",4,"reti   ",4,"push   ",3,"pop    ",4,"set1   ",4,"clr1   "
db 4,"tset1  ",4,"tclr1  ",4,"and1   ",3,"or1    ",4,"eor1   ",4,"not1   "
db 4,"mov1   ",4,"clrc   ",4,"setc   ",4,"notc   ",4,"clrv   ",4,"clrp   "
db 4,"setp   ",2,"ei     ",2,"di     ",3,"nop    ",4,"sleep  ",4,"stop"

;SpcOpcNamesExplantory:
;db 3,"move into",3,"add with carry",3,"subtract with borrow",3,"compare",3,"and",2,"or"
;db 3,"exclusive or",3,"increment",3,"decrement",3,"arithmetic shift left",3,"LSR    ",3,"roll left"
;db 3,"roll right",3,"exchange nybbles",4,"move word",4,"increment word",4,"decrement word",4,"add word"
;db 4,"subtract word",4,"compare word",3,"multiply",3,"divide",3,"adjust decimal addition",3,"adjust decimal subtraction"
;db 3,"branch",3,"branch if equal",3,"branch if unequal",3,"branch if carry set",3,"branch if carry clear",3,"branch if overflow set"
;db 3,"branch if overflow clear",3,"branch if sign set",3,"branch if sign clear",3,"branch if bit set",3,"branch if bit clear",4,"compare to A and branch if unequal"
;db 4,"decrement and branch if not zero",3,"jmp short",4,"jump long",4,"call",5,"user page call",5,"table call",3,"break"
;db 3,"return",4,"return from interrupt",4,"push onto stack",3,"pop from stack",4,"set bit",4,"clear bit"
;db 4,"test and set bits with A",4,"test and clear bits with A",4,"and carry with",3,"or carry with",4,"exclusive or carry with",4,"not"
;db 4,"move into carry",4,"clear carry",4,"set carry",4,"complement carry",4,"clear overflow and half-carry",4,"clear direct page flag"
;db 4,"set direct page flag",2,"enable interrupts",2,"disable interrupts",3,"no operation",4,"sleep until interrupt",4,"stop execution"
;SpcOpcNamePrepositions:
;into from with

SpcOprNames:
;Special codes:  (all other characters are literal)
; 0 = end of operand string
; 1 = byte value
; 2 = word value
; 3 = value in upper nybble of opcode (tcall)
; 4 = membit
; 5 = relative jump
; 6 = value in upper three bits of opcodes
; 7 = dp address in first operand
; 8 = dp address in second operand
.HighestCode    equ 8
.None:
.A:             db "A",0
.X:             db "X",0
.Y:             db "Y",0
.YA:            db "YA",0
.Sp:            db "Sp",0
.Psw:           db "Psw",0
.C:             db "C",0
.Im:            db 1,0
.DpO:           db "[",7,"]",0
.DpO2:          db "[",8,"]",0
.DpX:           db "[X]",0
.DpXInc:        db "[X+]",0
.DpOX:          db "[",7,"+X]",0
.DpY:           db "[Y]",0
.DpOY:          db "[",7,"+Y]",0
.AbsO:          db "[",2,"]",0
.AbsOX:         db "[",2,"+X]",0
.AbsOY:         db "[",2,"+Y]",0
.InDpOX:        db "[[",7,"+X]]",0
.YInDpO:        db "[[",7,"]+Y]",0
.InAbsOX:       db "[",2,"+X]",0
.MemBit:        db "[",4,"]",0
.MemBitN:       db "[-",4,"]",0
.DpOBit:        db "[",7,".",6,"]",0
.JumpRel:       db 5,0
.JumpAbs:       db 2,0
.OpcH:          db 3,0
.UsrPage:       db 1,0

align 4
.Table:
dd .None   ,.A      ,.X      ,.Y      ,.YA     ,.Sp     ,.Psw    ,.Im
dd .DpX    ,.DpXInc ,.DpY    ,.DpO    ,.DpOX   ,.DpOY   ,.AbsO   ,.AbsOX
dd .AbsOY  ,.InDpOX ,.YInDpO ,.JumpRel,.JumpAbs,.MemBit ,.MemBitN,.DpOBit
dd .InAbsOX,.OpcH   ,.UsrPage,.C      ,.DpO2

SpcFlagNames:
db "n",SpcFlag.N
db "v",SpcFlag.V
db "p",SpcFlag.P
db "b",SpcFlag.B
db "h",SpcFlag.H
db "i",SpcFlag.I
db "z",SpcFlag.Z
db "c",SpcFlag.C

;============================================================

%include "spcemu.asm"
%include "dspemu.asm"
%include "disasm.asm"
%include "midi.asm"
%include "sound.asm"
%include "gui.asm"
%include "spcview.asm"

;============================================================
section .data

SpcFileName:        dd 0
DspFileHandle:      dd 0

DspBufferPointer:   dd 0
DspBufferWritten:   dd 0
NoteBufferEnd:      dd 0    ;last pointer in buffer to hold next note
NoteBufferPos:      dd 0    ;current position in buffer
NoteBufferPrevPos:  dd 0    ;previous position in buffer
NoteBufferPlayed:   dd 0    ;portion of buffer played so far, number of new
                            ;notes is NoteBufferPointer - NoteBufferPlayed
CycleCount:         dd 0    ;actual time (if timed)
CycleCountPrev:     dd 0
CyclesEmulated:     dd 0    ;number cycles emulated so far
CyclesToEmulate:    dd -1   ;CyclesToEmulateAhead
                            ;time limit that emulation loop must not excede
                            ;set to -1 for maximum speed (not timed)
OpsEmulated:        dd 0    ;total number of instructions executed
CyclesPerInterrupt: dd CyclesPerPcInterrupt
InterruptAccum:     dd 0

SampleInfoEntries:  dd 1    ;total number of entries in sample info list (must always be at least one)

PlayOptions:
.Sounds             equ 1   ;make sounds while playing
.Timed              equ 2   ;play real speed, otherwise as fast as possible
.Record             equ 4   ;record or pause
.MidiLog            equ 8   ;log to a MIDI file
.DspLog             equ 16  ;log DSP writes to a file
                    db .Sounds | .MidiLog | .DspLog | .Timed
ActiveChannels:     db 255

;absolute file positions in the SPC and lengths of each piece
align 4
SpcFilePositions:
dd SpcRegs,25h,7                 ;Registers
dd SpcRam ,100h,65536            ;RAM
dd DspRam ,10100h,128            ;DSP
dd SpcEram,101C0h,64             ;Extra RAM
dd 0

;absolute file positions in the savestate and lengths of each piece
ZstFilePositions:
dd SpcRam ,199699,65536         ;RAM
dd SpcRegs,265251,48            ;Registers
dd SpcEram,265299,64            ;Extra RAM
dd DspRam ,266623,128           ;DSP memory (128 registers)
dd LoadSpcState.SpcSaved,36,1   ;flags whether savestate has sound emulation
dd 0

DefaultSampleInfo:
    dd 0    ;sample length (not bytes)
    dd 0    ;length of loop portion
    dd 0    ;addition checksum of both sample and loop
    db 138  ;nearest half-semitone when played at actual frequency (69*2 or A440 doubled)
    db 0    ;MIDI instrument or patch (0-127)
    db 64   ;volume
    db 0    ;bank for MIDI sequencers
    db 0    ;interpolation, simply for possible compatibility with SpcTool
    db "Default",0

align 2
PitchList:          incbin "pitchlst.dat"   ;130 words of nearest pitches, the first and last entries are 0 and 16384
PitchTableRle:      incbin "pitchtbl.dat"   ;word rle encoded
SbFreqTable:        incbin "fnumber.dat"

GuiPalette:         incbin "rainbow.pal"

;==============================
Text:
.Title:             db "Spc2Midi ",ProgVersion," - DSP to MIDI converter. (c)2000 PeekinSot",13,10
                    db 10,0
.About:             db "Now that the hard part is nearly finished, an SPC-700 emulation core faster",13,10
                    db "than that of SpcPlay, it's time to start working on the next step. That was",13,10
                    db "some time-consuming experience...but maybe I learned a few things.",13,10
                    db 10
                    db "This version does practically nada except load a savestate or spc (ZST,ZMV,SPC)",13,10
                    db "and emulate the processor. It is still a long way from decent MIDI exporting.",13,10
                    db 10
                    db "Usage:  Spc2Midi filename.ext",13,10,0
.UnverifiedOpcode:  db "##    @#### <########> - Unverified opcode. Emulation may procede incorrectly.",13,10,0
.DspWrite:          db "##:## @#### <########> - DSP written to.",13,10,0
.DspRead:           db "##:## @#### <########> - DSP read from.",13,10,0
.DspWriteIllegal:   db "##:## @#### <########> - Write to invalid DSP register.",13,10,0
.RegWriteIllegal:   db "##:## @#### <########> - Write to read-only register.",13,10,0
.RegReadIllegal:    db "##:## @#### <########> - Read from write-only register.",13,10,0
.FileTypeBad:       db "Unrecognized filetype. Spc2Midi can only load SPC captures and savestates.",13,10
                    db "Valid filetypes are SPC, ZST, or ZMV.",13,10,0
.FileOpenError:     db "Could not open the given file. Check the spelling and path.",13,10,0
.SpcReadError:      db "File does not contain SPC data or is corrupt.",13,10
                    db "Remember that ZSNES does not save SPC data if sound is disabled.",13,10,0
.DspOutputError:    db "Could not open the DSP output file.",13,10,0
.NumberTable:       db "0123456789ABCDEF"
.NumberFillChar:    db ''
;.SelectorError:    db "Could not access zero-based memory.",0
.LoadingFile:       db "Loading CPU and DSP state.",0
.InitializingSound: db "Initializing sound card for FM music.",0
.BufferingSamples:  db "Buffering BRR sound samples.",0
.LoadingSettings:   db "Loading settings from xxxxxxxx.xxx",0
.Ready:             db "Ready for emulation. Press a key...",13,10
                    db 10
                    db "During play:",13,10
                    db "  1-8    Mute/Enable",13,10
                    db "  0      Mute all",13,10
                    db "  9      Enable all",13,10
                    db "  Space  Pause",13,10
                    db "  BkSpc  Restart",13,10
                    db "  Other  Quit"
%if SpcDebugMode > 0
                    db 13,10,10
%endif
                    db 0
.SampleInfoTitles:  db "Sample  Offset       Length       Checksum",13,10,0
.SampleInfo:        db "###     ##### #####  ##### #####  ########",13,10,0

SoundDumpHeader:    db "RIFF"
                    dd 4+24+8+SamplesBuffer_Size
                    db "WAVEfmt "
                    dd 16
                    dw 1,1          ;PCM/Mono
                    dd 11025,22050  ;much lower than actual rate (32khz)
                    dw 1,16         ;block alignment/16bit audio
                    db "data"
                    dd SamplesBuffer_Size
SoundDumpHeader_Size equ $-SoundDumpHeader
SoundDumpFileName:  db "out.wav",0

%if SpcDebugDspOutput
DspFileName:        db "out.dsp",0
%endif
