;Spc2Midi - Display formats
;============================================================

section .bss
; Offscreen pixel buffer which the display routine has full reign to draw in
; The 32k buffer is then later copied to the screen.
SpcViewWndBuffer_Size   equ 128*256
SpcViewWndBuffer:       resb SpcViewWndBuffer_Size

; A table for the display format to hold whatever various values it needs to
; The table is twice the width of the window (256*2)
SpcViewWndTable_Size    equ 512
SpcViewWndTable:        resb SpcViewWndTable_Size

; Bit table to keep track of values in the above have changed
SpcViewWndChanges_Size  equ 256
SpcViewWndChanges       resb SpcViewWndChanges_Size

;
section .data
align 4
SpcViewWnd:
.Handler:       dd SpcViewWndDrawers.Hns
.ScreenCopier:  dd SpcViewWndCopy.PartialV
.Pos:           dd 0    ;row or column
.PrevVoiceNotes:db 0,0,0,0,0,0,0,0,0, 0,0,0

section .text
;
SpcViewWndSetMode:
    ;select case
.ClearTables:
    mov edi,SpcViewWndBuffer
    mov ecx,(SpcViewWndBuffer_Size+SpcViewWndTable_Size)/4
.Clear:
    cld
    xor eax,eax
    rep stosd
    mov ecx,SpcViewWndChanges_Size/4
  %if SpcViewWndBuffer+SpcViewWndBuffer_Size+SpcViewWndTable_Size != SpcViewWndChanges
    %error SpcViewWndChanges must come after SpcViewWndTable
  %endif
    dec eax
    rep stosd
    ret
.ClearTable:
    mov edi,SpcViewWndTable
    mov ecx,SpcViewWndTable_Size/4
    jmp short .Clear

;
SpcViewWndDrawers:
; Various display formats to view the notes in

;
; Horizontal Note Sheet
.Hns:
    call .NsReadBuffer
    mov ebx,[SpcViewWnd.Pos]
    mov ecx,64                  ;128/2
    mov edx,ebx
    inc edx                     ;row = (row + 1) and 255
    bts dword [SpcViewWndChanges],ebx
    and edx,255
    lea edi,[SpcViewWndBuffer+ebx]
    mov [SpcViewWnd.Pos],edx
.HnsNextBar:
    mov byte [edi],0
    mov byte [edi+256],1
    add edi,512
    dec ecx
    jnz .HnsNextBar

;Draw dashed song position bar
    mov ecx,64                  ;128/2
    bts dword [SpcViewWndChanges],edx
    lea edi,[SpcViewWndBuffer+edx+256]
.HnsNextHightlight:
    mov byte [edi],4
    add edi,512
    dec ecx
    jnz .HnsNextHightlight

    mov ecx,7
    lea edi,[SpcViewWndBuffer+ebx]
.HnsNextVoice:
    bt dword [ActiveChannels],ecx
    jnc .HnsSkipVoice
    mov edx,[SpcViewWndTable+64+ecx*4]  ;get note decay
    cmp edx,9
    jb .HnsSkipVoice
    je .NoMoreDecay
    dec edx
    mov [SpcViewWndTable+64+ecx*4],edx  ;set new note decay
.NoMoreDecay:
    mov ebx,[SpcViewWndTable+ecx*4]     ;get note value
    mov eax,ecx
    xor ebx,127                 ;reverse value so that notes are not upside-down
    shl eax,4                   ;*16
    shl ebx,8                   ;*256, width of buffer
    sub eax,ecx                 ;-[voice number]
    lea eax,[eax*2+edx+16]      ;add voice's note decay (skipping first 16 gui colors) to get rainbow shade
    mov [edi+ebx],al            ;color formula = voice * 31 + 30  OR  ((voice * 16) - voice) * 2 + 30
.HnsSkipVoice:
    dec ecx
    jns .HnsNextVoice
    ret

;
; Vertical Note Sheet
.Vns:
    call .NsReadBuffer
    cld
    mov ebx,[SpcViewWnd.Pos]
    bts dword [SpcViewWndChanges],ebx
    mov edx,ebx
    shl ebx,8                   ;*256
    inc edx                     ;row = (row + 1) and 127
    lea edi,[SpcViewWndBuffer+ebx]
    and edx,127
    mov ecx,256/4
    mov [SpcViewWnd.Pos],edx
    mov eax,00080008h           ;gray/black/gray/black
    rep stosd

    mov ecx,7
.VnsNextVoice:
    mov ebx,[SpcViewWndTable+ecx*4];get voice's note
    test ebx,ebx                ;test if null
    jz .VnsSkipVoice
    mov eax,ecx
    shl eax,1
    add eax,byte 32
    mov ah,al
    mov [edi-256 +ebx*2-2],ax
.VnsSkipVoice:
    dec ecx
    jns .VnsNextVoice
    ret

;
; Note Sheet buffer reader
.NsReadBuffer:
    mov esi,[NoteBufferPrevPos]
    and esi,NoteBufferSize-1
    jmp short .NsFirst
.NsNextNote:
    mov ebx,[NoteBuffer+esi+4]  ;get note number/sample/volume
    test bl,1
    jnz .NsNotKeyOn

    movzx edi,byte [NoteBuffer+esi+6];get sample number
    mov eax,ebx
    mov edi,[DspSampleTable.AtrsIdx+edi*4]
    shr ebx,4                   ;get voice/channel
    shl edi,SampleAtrs_Shift
    sub ah,byte 207             ;reduce note value default from 208 to 1
    and ebx,111b                ;voice/channel 0-7
    add ah,[SampleAtrs.HalfSemitone+edi]
    shr eax,8+1                 ;get half-semitone and /2 to get note number

    and eax,01111111b           ;mask note number 0-127
    mov [SpcViewWndTable+ebx*4],eax;set voice's note
    mov dword [SpcViewWndTable+64+ebx*4],14;set voice'initial decay
    jmp short .NsAfterKeyChange
.NsNotKeyOn:
    test bl,2
    jnz .NsNotKeyOff
    shr ebx,4                   ;get voice/channel
    and ebx,111b                ;voice/channel 0-7
    mov dword [SpcViewWndTable+64+ebx*4],0;set decay to zero
.NsNotKeyOff:
.NsAfterKeyChange:
    add esi,byte 8              ;next note in buffer
    and esi,NoteBufferSize-1
.NsFirst:
    cmp [NoteBufferPos],esi
    jne .NsNextNote
    ret

;
; Vertical Bars
.Vb:
;Decrement volume of all bars
    mov ecx,254                 ;start from last
.VbNext:
    mov al,[SpcViewWndTable+ecx]
    test al,al
    jz .VbBaseLine
    cmp [SpcViewWndTable+256+ecx],al
    je .VbBaseLine
    bts dword [SpcViewWndChanges],ecx
    ;inc ecx
    ;bts dword [SpcViewWndChanges],ecx
    sub al,13
    jnc .VbNoUnderflow
    xor al,al
    mov [SpcViewWndTable+256+ecx],al
.VbNoUnderflow:
    mov [SpcViewWndTable+ecx],al
.VbBaseLine:
    sub ecx,byte 2
    jns .VbNext

;Read through note buffer to find any new notes
    mov esi,[NoteBufferPrevPos]
    and esi,NoteBufferSize-1
    jmp short .VbFirst
.VbNextNote:
    mov eax,[NoteBuffer+esi+4]  ;get note number/sample/volume
    test al,1
    jnz .VbNotKeyOn
    mov ecx,eax
    shr eax,4
    shr ecx,8
    and eax,111b                ;voice/channel
    and ecx,11111110b           ;note number
    mov [SpcViewWndTable+1+ecx],al;set voice (bar color)
    mov [SpcViewWnd.PrevVoiceNotes+eax],cl
    mov [SpcViewWndTable+256+ecx],byte 64
    mov [SpcViewWndTable+ecx],byte 64
    bts dword [SpcViewWndChanges],ecx
    jmp short .VbAfterKeyChange
    ;inc ecx
    ;bts dword [SpcViewWndChanges],ecx
;    shr eax,24+1                ;get volume and /2
;    cmp al,16                   ;check that it is at last 10 pixels tall
;    jae .VolumeOk
;    mov al,16
;.VolumeOk:
.VbNotKeyOn:
    test al,2
    jnz .VbNotKeyOff
    shr eax,4
    and eax,111b                ;voice/channel
    movzx ecx,byte [SpcViewWnd.PrevVoiceNotes+eax]
    mov [SpcViewWndTable+256+ecx],byte 0
    bts dword [SpcViewWndChanges],ecx
.VbNotKeyOff:
.VbAfterKeyChange:
    add esi,byte 8              ;next note in buffer
    and esi,NoteBufferSize-1
.VbFirst:
    cmp [NoteBufferPos],esi
    jne .VbNextNote

;Redraw any changes in buffer
    mov ecx,254
.VbCheckNextBar:
    bt dword [SpcViewWndChanges],ecx
    jnc .VbNoRedraw
    lea edi,[SpcViewWndBuffer+ecx];get destination in buffer
    mov eax,[SpcViewWndTable+ecx];get bar height and color
    movzx ebx,al                ;get bar height
    ;and ebx,127                 ;limit to 0-127
    mov edx,ebx
    not ebx
    and ebx,127                 ;get 128 - bar height
    ;shl ah,4
    ;add ah,16
.VbNextBarTop:
    mov word [edi],0008h
    add edi,256
    dec ebx
    jns .VbNextBarTop
    test edx,edx
    jz .VbNoRedraw
    shl ah,1
    add ah,32
    mov al,ah                   ;make color word
.VbNextBarBtm:
    mov [edi],ax
    add edi,256
    dec edx
    jnz .VbNextBarBtm
    cmp edi,SpcViewWndBuffer+32768+256
    jb .ok
    int3;
.ok:
.VbNoRedraw:
    sub ecx,byte 2
    jns .VbCheckNextBar
    ret

;
; Temporary function until the GUI is completed.
SpcViewWndCopy:
.Complete:
    cld
    mov ebx,128
    mov esi,SpcViewWndBuffer
    mov edi,Screen.Dest
.NextPixelRow:
    mov ecx,256/4
    rep movsd
    add edi,byte 64
    dec ebx
    jnz .NextPixelRow
    ret

.PartialV:
    push dword 128
    mov ebx,SpcViewWndBuffer
    mov edx,Screen.Dest
.NextRow:
    xor ecx,ecx
.NextStrip:
    mov eax,[SpcViewWndChanges+ecx]
    mov esi,ebx
    mov edi,edx
.Next:
    test eax,15
    jz .SkipBlit
    movsd
    shr eax,4
    jnz .Next
    jmp short .AfterStripBlit
.SkipBlit:
    add esi,byte 4
    add edi,byte 4
    shr eax,4
    jnz .Next
.AfterStripBlit:
    add ebx,byte 32
    add edx,byte 32
    add ecx,byte 4
    cmp ecx,32
    jb .NextStrip
    add edx,byte 64
;    cmp edi,0B0000h
;    jb .ok
;    int3;
;.ok:
    dec dword [esp]
    jnz .NextRow
    pop ecx

;Reset change flags
    cld
    xor eax,eax
    mov ecx,SpcViewWndChanges_Size/4
    mov edi,SpcViewWndChanges
    rep stosd
    ret

;
; Temporary function until the GUI is completed.
; This does not belong in here.

;(esi=palette)
SetPalette:
    mov ecx,256
;(esi=palette, ecx=count)
.GivenCount:
    xor eax,eax
;(esi=palette, eax=starting reg, ecx=count)
.GivenIndexes:
    mov edx,3C8h
    out dx,al

    inc edx
    and ecx,511
    lea ecx,[ecx*2+ecx]
    cld
    rep outsb

;    mov edi,Screen.Dest+(180*320)
;    xor al,al
;.NextDot:
;    mov [edi+320],al
;    stosb
;    inc al
;    jnz .NextDot
    ret

;
; Temporary function until the GUI is completed.
; This does not belong in here.

ShowActiveChannels:
    mov edi,Screen.Dest+(129*320)
    xor ecx,ecx
    mov ebx,1E1E1E1Eh
.NextChannel:
    bt dword [ActiveChannels],ecx
    mov eax,ebx
    jc .ActiveChannel
    xor eax,eax
.ActiveChannel:
    mov [edi],eax
    mov [edi+320],eax
    shr eax,8
    mov [edi+4],eax
    mov [edi+324],eax
.SkippedChannel:
    add edi,byte 8
    inc ecx
    add ebx,1E1E1E1Eh
    cmp ecx,8
    jb .NextChannel
    ret
